﻿using UnityEngine;
using System.Collections;

// [Unity] Procedural Cave Generation (E02. Marching Squares)
// by Sebastian Lague.
public class MeshGenerator : MonoBehaviour
{
	public SquareGrid SquareGrid_;

	void OnDrawGizmos ()
	{
		if (SquareGrid_ != null) {
			for (int x = 0; x < SquareGrid_.squares.GetLength (0); x++) {
				for (int y = 0; y < SquareGrid_.squares.GetLength (1); y++) {
					Gizmos.color = SquareGrid_.squares [x, y].TopLeft.Active ? Color.black : Color.white;
					Gizmos.DrawCube (SquareGrid_.squares [x, y].TopLeft.Position, Vector3.one * .4f);

					Gizmos.color = SquareGrid_.squares [x, y].TopRight.Active ? Color.black : Color.white;
					Gizmos.DrawCube (SquareGrid_.squares [x, y].TopRight.Position, Vector3.one * .4f);

					Gizmos.color = SquareGrid_.squares [x, y].BottomRight.Active ? Color.black : Color.white;
					Gizmos.DrawCube (SquareGrid_.squares [x, y].BottomRight.Position, Vector3.one * .4f);

					Gizmos.color = SquareGrid_.squares [x, y].BottomLeft.Active ? Color.black : Color.white;
					Gizmos.DrawCube (SquareGrid_.squares [x, y].BottomLeft.Position, Vector3.one * .4f);

					// Mid point
					Gizmos.color = Color.grey;
					Gizmos.DrawCube (SquareGrid_.squares [x, y].CentreTop.Position, Vector3.one * .15f);
					Gizmos.DrawCube (SquareGrid_.squares [x, y].CentreRight.Position, Vector3.one * .15f);
					Gizmos.DrawCube (SquareGrid_.squares [x, y].CentreBottom.Position, Vector3.one * .15f);
					Gizmos.DrawCube (SquareGrid_.squares [x, y].CentreLeft.Position, Vector3.one * .15f);

				}
			}
		}
	}

	public void GenerateMesh (int[,] map, float squareSize)
	{
		SquareGrid_ = new SquareGrid (map, squareSize);
	}

	// 2D array of Squares.
	public class SquareGrid
	{
		public Square[,] squares;

		public SquareGrid (int[,] map, float squareSize)
		{
			int nodeCountX = map.GetLength (0);
			int nodeCountY = map.GetLength (1);
			float mapWidth = nodeCountX * squareSize;
			float mapHeight = nodeCountY * squareSize;

			ControlNode[,] controlNodes = new ControlNode[nodeCountX, nodeCountY];

			for (int x = 0; x < nodeCountX; x++) {
				for (int y = 0; y < nodeCountY; y++) {
					// -mapWidth / 2 で原点にする。
					Vector3 position = new Vector3 (-mapWidth / 2 + x * squareSize + squareSize / 2, 0, -mapHeight / 2 + y * squareSize + squareSize / 2);
					bool active = (map [x, y] == 1);
					controlNodes [x, y] = new ControlNode (position, active, squareSize);
				}
			}

			// X座標の頂点が、2つなら四角形は1つ、3つなら四角形は2つ
			// つまり、頂点ノード数-1が四角形の数になるため、-1している
			squares = new Square[nodeCountX - 1, nodeCountY - 1];

			for (int x = 0; x < nodeCountX - 1; x++) {
				for (int y = 0; y < nodeCountY - 1; y++) {
					squares [x, y] = new Square (
						controlNodes [x, y + 1],
						controlNodes [x + 1, y + 1],
						controlNodes [x + 1, y],
						controlNodes [x, y] // ボトムレフトが基準
					);
				}
			}
		}
	}

	// 頂点4つで四角形を保持する。
	// ControlNodeは上と右に隣接するノード情報を保持する。
	// C = ControlNode, N = Nodeとすると以下の四角形を保持することになる。
	// C----N----C
	// |         |
	// |         |
	// N         N
	// |         |
	// |         |
	// C----N----C
	public class Square
	{
		public ControlNode TopLeft, TopRight, BottomRight, BottomLeft;
		public Node CentreTop, CentreRight, CentreBottom, CentreLeft;

		public Square (ControlNode topLeft, ControlNode topRight, ControlNode bottomRight, ControlNode bottomLeft)
		{
			TopLeft = topLeft;
			TopRight = topRight;
			BottomRight = bottomRight;
			BottomLeft = bottomLeft;

			// エイリアス.
			CentreTop = TopLeft.Right;
			CentreRight = BottomRight.Above;
			CentreBottom = BottomLeft.Right;
			CentreLeft = BottomLeft.Above;
		}
	}

	// コントロールノードは、隣接するノードを2つ持つ
	// 別に上下左右に隣接する４つのノードを持ってもよいが、コントロールノードが
	// 格子状に並んだ場合、上と右のノードがわかればよいそのため２つノードを持つ
	public class ControlNode : Node
	{
		public bool Active;
		public Node Above, Right;

		// 自分の座標がpos、半径がsquareSize/2となる。
		// 自身のposから、Z軸に、自分の半径分、上(=Abobe)移動したところに、ノードを作成
		// 自身のposから、X軸に、自分の半径分、右(=Abobe)移動したところに、ノードを作成
		public ControlNode (Vector3 pos, bool active, float squareSize) : base (pos)
		{
			Active = active;
			Above = new Node (Position + Vector3.forward * squareSize / 2f); // forward is Z axis
			Right = new Node (Position + Vector3.right * squareSize / 2f); // right is X axis
		}

	}

	// ノードは頂点情報。
	public class Node
	{
		public Vector3 Position;
		public int VertexIndex = -1;

		public Node (Vector3 pos)
		{
			Position = pos;
		}
	}
}
